import pandas as pd
from sklearn.preprocessing import MinMaxScaler, QuantileTransformer, StandardScaler
from ..utils.data import get_numeric_categorical_columns
from ..algorithms.linear_regression import lr_get_mse


def apply_Min_Max_Normalization(df, target, config={}, eval_method=lr_get_mse):

    numeric_columns, categorical_columns = get_numeric_categorical_columns(df)
    target_column = numeric_columns.loc[:, target]
    nontarget_columns = numeric_columns.loc[:, numeric_columns.columns != target]

    normalizer = MinMaxScaler().fit(nontarget_columns)
    normalized_arr = normalizer.transform(nontarget_columns)
    normalized_df = pd.DataFrame(normalized_arr, columns=nontarget_columns.columns)

    processed_dataset = pd.concat(
        [normalized_df, categorical_columns, target_column], axis=1
    )

    try:
        dataframe, mse = eval_method(processed_dataset, target)
        return processed_dataset, mse
    except:
        return processed_dataset, float("inf")


def apply_Quantile_Normalization(
    df, target, config={"n_quantiles": 10, "random_state": 0}, eval_method=lr_get_mse
):

    numeric_columns, categorical_columns = get_numeric_categorical_columns(df)
    target_column = numeric_columns.loc[:, target]
    nontarget_columns = numeric_columns.loc[:, numeric_columns.columns != target]

    n_quantiles = config["n_quantiles"]
    random_state = config["random_state"]

    normalizer = QuantileTransformer(n_quantiles=n_quantiles, random_state=random_state)
    normalized_arr = normalizer.fit_transform(nontarget_columns)
    normalized_df = pd.DataFrame(normalized_arr, columns=nontarget_columns.columns)

    processed_dataset = pd.concat(
        [normalized_df, categorical_columns, target_column], axis=1
    )

    try:
        dataframe, mse = eval_method(processed_dataset, target)
        return processed_dataset, mse
    except:
        return processed_dataset, float("inf")


def apply_Z_Score_Normalization(df, target, config={}, eval_method=lr_get_mse):

    numeric_columns, categorical_columns = get_numeric_categorical_columns(df)
    target_column = numeric_columns.loc[:, target]
    nontarget_columns = numeric_columns.loc[:, numeric_columns.columns != target]
    normalizer = StandardScaler().fit(nontarget_columns)
    normalized_arr = normalizer.transform(nontarget_columns)
    normalized_df = pd.DataFrame(normalized_arr, columns=nontarget_columns.columns)

    processed_dataset = pd.concat(
        [normalized_df, categorical_columns, target_column], axis=1
    )

    try:
        dataframe, mse = eval_method(processed_dataset, target)
        return processed_dataset, mse
    except:
        return processed_dataset, float("inf")



import pandas as pd


def get_numeric_categorical_columns(df):
    categorical_columns = []
    numeric_columns = []
    for col in df.columns:
        if df[col].map(type).eq(str).any():  # check if there are any strings in column
            categorical_columns.append(col)
        else:
            numeric_columns.append(col)

    # create two DataFrames, one for each data type
    return pd.DataFrame(df[numeric_columns]), pd.DataFrame(df[categorical_columns])



import pandas as pd
import numpy as np
from sklearn.feature_selection import SelectKBest, chi2
from sklearn.feature_selection import VarianceThreshold
from ..utils.data import get_numeric_categorical_columns
from ..algorithms.linear_regression import lr_get_mse


def apply_Missing_Ratio_Feature_Selection(
    df, target, config={"threshold": 0.2}, eval_method=lr_get_mse
):

    target_column = df.loc[:, target]
    nontarget_columns = df.loc[:, df.columns != target]

    missing_series = nontarget_columns.isnull().sum() / nontarget_columns.shape[0]

    missing_stats = pd.DataFrame(missing_series).rename(
        columns={"index": "feature", 0: "missing_fraction"}
    )

    # Sort with highest number of missing values on top
    missing_stats = missing_stats.sort_values("missing_fraction", ascending=False)

    # Find the columns with a missing percentage above the threshold
    record_missing = (
        pd.DataFrame(missing_series[missing_series > config["threshold"]])
        .reset_index()
        .rename(columns={"index": "feature", 0: "missing_fraction"})
    )

    to_drop = list(record_missing["feature"])
    to_keep = set(df.columns) - set(to_drop)

    processed_dataset = pd.concat([df[list(to_keep)], target_column], axis=1)

    try:
        dataframe, mse = eval_method(processed_dataset, target)
        return processed_dataset, mse
    except:
        return processed_dataset, float("inf")


def apply_K_Best_Feature_Selection(df, target, config={"k": 5}, eval_method=lr_get_mse):
    numeric_columns, categorical_columns = get_numeric_categorical_columns(df)
    target_column = numeric_columns.loc[:, target]
    nontarget_columns = numeric_columns.loc[:, numeric_columns.columns != target]

    # Extract +ve columns
    lsv = list(nontarget_columns.lt(0).sum().values)
    lis = list(nontarget_columns.lt(0).sum().index)
    to_remove = []
    for i in range(0, len(lsv)):
        if lsv[i] > 0:
            to_remove.append(lis[i])
    lis = list(filter(lambda x: x not in to_remove, lis))

    selection = nontarget_columns
    if len(lis) > 0:
        filtered_df = nontarget_columns[lis]
        try:
            selector = SelectKBest(chi2, k=config["k"])
            selector.fit(filtered_df, target_column)

            cols = selector.get_support(indices=True)
        except:
            return df, float("inf")
        selection = filtered_df.iloc[:, cols]

    processed_dataset = pd.concat(
        [selection, categorical_columns, target_column], axis=1
    )

    try:
        dataframe, mse = eval_method(processed_dataset, target)
        return processed_dataset, mse
    except:
        return processed_dataset, float("inf")


def apply_Variance_Based_Feature_Selection(
    df, target, config={"threshold": 0}, eval_method=lr_get_mse
):
    numeric_columns, categorical_columns = get_numeric_categorical_columns(df)
    target_column = numeric_columns.loc[:, target]
    nontarget_columns = numeric_columns.loc[:, numeric_columns.columns != target]

    selector = VarianceThreshold(threshold=config["threshold"])
    try:
        selector.fit(nontarget_columns, target_column)
        cols = selector.get_support(indices=True)
        selection = nontarget_columns.iloc[:, cols]
    except Exception as e:
        selection = nontarget_columns

    processed_dataset = pd.concat(
        [selection, categorical_columns, target_column], axis=1
    )

    try:
        dataframe, mse = eval_method(processed_dataset, target)
        return processed_dataset, mse
    except:
        return processed_dataset, float("inf")



import pandas as pd
import numpy as np
from sklearn.impute import KNNImputer, SimpleImputer
from ..utils.data import get_numeric_categorical_columns
from ..algorithms.linear_regression import lr_get_mse


def apply_KNN_Imputation(
    df, target, config={"n_neighbors": 2, "weights": "uniform"}, eval_method=lr_get_mse
):

    numeric_columns, categorical_columns = get_numeric_categorical_columns(df)

    n_neighbors = config["n_neighbors"]
    weights = config["weights"]

    imputer = KNNImputer(n_neighbors=n_neighbors, weights=weights)
    imputed_arr = imputer.fit_transform(numeric_columns)
    imputed_df = pd.DataFrame(imputed_arr, columns=numeric_columns.columns)

    processed_dataset = pd.concat([imputed_df, categorical_columns], axis=1)

    try:
        dataframe, mse = eval_method(processed_dataset, target)
        return processed_dataset, mse
    except:
        return processed_dataset, float("inf")


def apply_Simple_Imputation(df, config={}):

    strategy = config["strategy"]
    missing_values = config["missing_values"]

    imputer = SimpleImputer(strategy=strategy, missing_values=missing_values)
    imputed_arr = imputer.fit_transform(df)
    imputed_df = pd.DataFrame(imputed_arr, columns=df.columns)

    return imputed_df


def apply_most_frequent_value_imputer(df, target, config={}, eval_method=lr_get_mse):
    config["strategy"] = "most_frequent"

    numeric_columns, categorical_columns = get_numeric_categorical_columns(df)
    config["missing_values"] = np.nan
    imputed_numeric_columns = apply_Simple_Imputation(numeric_columns, config)
    if len(categorical_columns.columns) > 0:
        config["missing_values"] = "NaN"
        imputed_categorical_columns = apply_Simple_Imputation(
            categorical_columns, config
        )
    else:
        imputed_categorical_columns = categorical_columns

    processed_dataset = pd.concat(
        [imputed_numeric_columns, imputed_categorical_columns], axis=1
    )

    try:
        dataframe, mse = eval_method(processed_dataset, target)
        return processed_dataset, mse
    except:
        return processed_dataset, float("inf")


def apply_mean_imputer(df, target, config={}, eval_method=lr_get_mse):
    config["strategy"] = "mean"

    numeric_columns, categorical_columns = get_numeric_categorical_columns(df)
    config["missing_values"] = np.nan
    imputed_numeric_columns = apply_Simple_Imputation(numeric_columns, config)

    processed_dataset = pd.concat(
        [imputed_numeric_columns, categorical_columns], axis=1
    )

    try:
        dataframe, mse = eval_method(processed_dataset, target)
        return processed_dataset, mse
    except:
        return processed_dataset, float("inf")


def apply_median_imputer(df, target, config={}, eval_method=lr_get_mse):
    config["strategy"] = "median"

    numeric_columns, categorical_columns = get_numeric_categorical_columns(df)
    config["missing_values"] = np.nan
    imputed_numeric_columns = apply_Simple_Imputation(numeric_columns, config)

    processed_dataset = pd.concat(
        [imputed_numeric_columns, categorical_columns], axis=1
    )

    try:
        dataframe, mse = eval_method(processed_dataset, target)
        return processed_dataset, mse
    except:
        return processed_dataset, float("inf")



import random


def epsilon(choices, q_vals, e):
    """
    choices is an array of possible actions and q_vals contains their corresponding q values
    this function chooses one and returns it
    """

    #     e = 0.8 #80% of the time function should choose highest quality
    num = random.randrange(0, 100) / 100
    if e >= num:  # choose based on highest quality
        #         print("Epsilon chose quality")
        max_q = max(q_vals)
        # accounts for potentially multiple choices with highest Q
        max_indices = [i for i, val, in enumerate(q_vals) if val == max_q]
        #         if multiple, randomly pick
        if len(max_indices) > 1:
            choice = random.randrange(0, len(max_indices))
            index = max_indices[choice]
        else:  # if just one, then return
            index = max_indices[0]
        return choices[index]
    else:  # choose randomly
        #         print("Epsilon chose random")
        index = random.randrange(0, len(choices))
        return choices[index]



import pandas as pd
import numpy as np
from scipy import stats
from sklearn.neighbors import LocalOutlierFactor
from ..utils.data import get_numeric_categorical_columns
from ..algorithms.linear_regression import lr_get_mse


def apply_MAD_Score_Based_Outlier_Detection(
    df, target, config={"threshold": 3.0, "ratio": 0.3}, eval_method=lr_get_mse
):
    numeric_columns, categorical_columns = get_numeric_categorical_columns(df)
    target_column = numeric_columns.loc[:, target]
    nontarget_columns = numeric_columns.loc[:, numeric_columns.columns != target]

    if df.isnull().values.any() > 0:
        return df, float("inf")

    median = nontarget_columns.apply(np.median, axis=0)

    # median_absolute_deviation = 1.4296 * \
    #     np.abs(nontarget_columns - median).apply(np.median, axis=0)
    median_absolute_deviation = stats.median_abs_deviation(nontarget_columns, scale=1)

    modified_z_scores = (nontarget_columns - median) / median_absolute_deviation

    outliers = nontarget_columns[np.abs(modified_z_scores) > config["threshold"]]

    to_drop = outliers[
        (outliers.count(axis=1) / outliers.shape[1]) > config["ratio"]
    ].index

    to_keep = set(nontarget_columns.index) - set(to_drop)

    if config["ratio"] == -1:
        filtered_df = nontarget_columns[
            ~(np.abs(modified_z_scores) > config["threshold"]).any(axis=1)
        ]
    else:
        filtered_df = nontarget_columns.loc[list(to_keep)]

    processed_dataset = pd.concat(
        [filtered_df, categorical_columns, target_column], axis=1
    )
    processed_dataset.dropna(inplace=True)

    try:
        dataframe, mse = eval_method(processed_dataset, target)
        return processed_dataset, mse
    except:
        return processed_dataset, float("inf")


def apply_Inter_Quantile_Range_Outlier_Detection(
    df, target, config={"ratio": 0.3}, eval_method=lr_get_mse
):
    numeric_columns, categorical_columns = get_numeric_categorical_columns(df)
    target_column = numeric_columns.loc[:, target]
    nontarget_columns = numeric_columns.loc[:, numeric_columns.columns != target]

    if nontarget_columns.isnull().values.any() > 0:
        return df, float("inf")

    Q1 = nontarget_columns.quantile(0.25)
    Q3 = nontarget_columns.quantile(0.75)

    inter_quatile_range = Q3 - Q1

    outliers = nontarget_columns[
        (
            (nontarget_columns < (Q1 - 1.5 * inter_quatile_range))
            | (nontarget_columns > (Q3 + 1.5 * inter_quatile_range))
        )
    ]

    to_drop = outliers[
        (outliers.sum(axis=1) / outliers.shape[1]) > config["ratio"]
    ].index

    to_keep = set(nontarget_columns.index) - set(to_drop)

    if config["ratio"] == -1:
        filtered_df = nontarget_columns[~(filter_criteria).any(axis=1)]
    else:
        filtered_df = nontarget_columns.loc[list(to_keep)]

    processed_dataset = pd.concat(
        [filtered_df, categorical_columns, target_column], axis=1
    )
    processed_dataset.dropna(inplace=True)

    try:
        dataframe, mse = eval_method(processed_dataset, target)
        return processed_dataset, mse
    except:
        return processed_dataset, float("inf")


def apply_Local_Factor_Outlier_Detection(
    df,
    target,
    config={"n_neighbors": 4, "contamination": 0.1, "threshold": 30},
    eval_method=lr_get_mse,
):
    numeric_columns, categorical_columns = get_numeric_categorical_columns(df)
    target_column = numeric_columns.loc[:, target]
    nontarget_columns = numeric_columns.loc[:, numeric_columns.columns != target]

    if nontarget_columns.isnull().values.any() > 0:
        return df, float("inf")

    clf = LocalOutlierFactor(n_neighbors=4, contamination=0.1)
    clf.fit_predict(nontarget_columns)

    LOF_scores = clf.negative_outlier_factor_

    k = config["threshold"]
    top_k_idx = np.argsort(LOF_scores)[-k:]
    top_k_values = [LOF_scores[i] for i in top_k_idx]

    filtered_df = nontarget_columns[LOF_scores < top_k_values[0]]

    processed_dataset = pd.concat(
        [filtered_df, categorical_columns, target_column], axis=1
    )
    processed_dataset.dropna(inplace=True)

    try:
        dataframe, mse = eval_method(processed_dataset, target)
        return processed_dataset, mse
    except:
        return processed_dataset, float("inf")



def apply_Exact_Duplicate_Detection(df, target, config={}, eval_method=lr_get_mse):
    processed_dataset = df.drop_duplicates()

    try:
        dataframe, mse = eval_method(processed_dataset, target)
        return processed_dataset, mse
    except:
        return processed_dataset, float("inf")



import numpy as np
from sklearn.linear_model import LinearRegression
from sklearn.metrics import mean_squared_error as mse
from sklearn.model_selection import train_test_split as tts


def lr_get_mse(df, target):
    try:
        xtrain, xtest, ytrain, ytest = tts(
            df.loc[:, df.columns != target],
            df.loc[:, target],
            test_size=0.3,
            random_state=69,
        )
        regr = LinearRegression()
        regr.fit(xtrain, ytrain)
        ypred = regr.predict(xtest)
        error = mse(ytest, ypred, squared=False)  # actually RMSE here, not MSE
        return df, error
    except:
        return df, float("inf")



"""A tool to automate preprocessing phase."""

import pandas as pd
from .preprocessing.outlier_detection import (
    apply_Local_Factor_Outlier_Detection,
    apply_MAD_Score_Based_Outlier_Detection,
)
from .preprocessing.feature_selection import (
    apply_K_Best_Feature_Selection,
    apply_Variance_Based_Feature_Selection,
    apply_Missing_Ratio_Feature_Selection,
)
from .preprocessing.imputation import (
    apply_most_frequent_value_imputer,
    apply_mean_imputer,
    apply_median_imputer,
)
from .preprocessing.normalization import (
    apply_Min_Max_Normalization,
    apply_Z_Score_Normalization,
    apply_Quantile_Normalization,
)
from .algorithms.linear_regression import lr_get_mse
from .utils.greedy import epsilon


def preprocmachine(df, target, goal="LinReg", gamma=0.8):

    numExpl = 300

    start = [lr_get_mse]
    imputation = [
        # apply_KNN_Imputation,
        apply_most_frequent_value_imputer,
        apply_mean_imputer,
        apply_median_imputer,
    ]
    outlier_detection = [
        apply_MAD_Score_Based_Outlier_Detection,
        # apply_Inter_Quantile_Range_Outlier_Detection,
        apply_Local_Factor_Outlier_Detection,
    ]
    normalization = [
        apply_Z_Score_Normalization,
        apply_Min_Max_Normalization,
        apply_Quantile_Normalization,
    ]
    feature_selection = [
        apply_Missing_Ratio_Feature_Selection,
        apply_K_Best_Feature_Selection,
        apply_Variance_Based_Feature_Selection,
    ]
    goal = [lr_get_mse]

    statefuncs = (
        start
        + imputation
        + outlier_detection
        + normalization
        + feature_selection
        + goal
    )
    statenames = []
    for i in statefuncs:
        statenames.append(i.__name__)

    # generate r table
    threshold1 = len(start) - 1
    threshold2 = threshold1 + len(imputation)
    threshold3 = threshold2 + len(outlier_detection)
    threshold4 = threshold3 + len(normalization)
    threshold5 = threshold4 + len(feature_selection)
    threshold6 = threshold5 + len(goal)

    r_table = []
    for i in range(len(statefuncs)):
        temp = []
        for j in range(len(statefuncs)):
            if i == len(statefuncs) - 1:
                temp.append(0)
            elif j == len(statefuncs) - 1:
                temp.append(1)
            elif (
                (i > threshold5 and j <= threshold6)
                or (i > threshold4 and j <= threshold5)
                or (i > threshold3 and j <= threshold4)
                or (i > threshold2 and j <= threshold3)
                or (i > threshold1 and j <= threshold2)
                or i == j
            ):
                temp.append(-1)
            else:
                temp.append(0)
        r_table.append(temp)

    #     initialize Q matrix

    q_table = []
    for i in range(len(statefuncs)):
        temp = []
        for j in range(len(statefuncs)):
            temp.append(0)
        q_table.append(temp)

    #     establish possible entry points
    entry_range = 0
    goal_state = len(statefuncs) - 1

    safecopy = pd.DataFrame.copy(df)

    for i in range(numExpl):

        # clear_output(wait=False)
        print("Starting exploration", i + 1, "/", numExpl)
        state = 0
        route = []  # will store state history for this exploration
        total_q = 0
        route.append(state)
        currExploreDF = pd.DataFrame.copy(df)
        currExploreDF, currExploreError = statefuncs[state](currExploreDF, target)

        while state != goal_state:
            possible_states = [i for i, val in enumerate(r_table[state]) if val >= 0]
            possible_qs = [q_table[state][i] for i in possible_states]
            prob = (i / numExpl) * 0.7
            next_state = epsilon(possible_states, possible_qs, prob)
            print("Moved from", state, "-->", next_state)

            # apply new state's proc to current df
            currExploreDF, newExploreError = statefuncs[next_state](
                currExploreDF, target
            )

            if currExploreError == float("inf") and newExploreError == float("inf"):
                newReward = -1
            elif currExploreError == float("inf"):
                newReward = 1
            elif newExploreError == float("inf"):
                newReward = -2
            else:
                newReward = (
                    (currExploreError - newExploreError) / currExploreError
                ) * 100
            reward = newReward
            currExploreError = newExploreError

            # find max Q value out of all possible actions from new state
            possible_states = [
                i for i, val in enumerate(r_table[next_state]) if val >= 0
            ]
            possible_qs = [q_table[next_state][i] for i in possible_states]
            max_q = max(possible_qs)

            # calculate new Q value for this state
            q_table[state][next_state] = (
                reward + gamma * max_q + r_table[state][next_state]
            )

            total_q += q_table[state][next_state]

            # move to next state
            state = next_state

            # record path
            route.append(state)

    print("Current Q table")
    for i in range(0, len(q_table)):
        print(q_table[i])

    # find best route
    best_route = []
    q_vals = []
    state = 0
    q = 0
    route.append(state)
    while state != len(statefuncs) - 1:
        possible_states = [i for i, val in enumerate(r_table[state]) if val >= 0]
        possible_qs = [q_table[state][i] for i in possible_states]
        # choose highest q and get corresponding state
        max_q = max(possible_qs)
        q_vals.append(max_q)
        q += max_q
        index = possible_qs.index(max_q)
        next_state = possible_states[index]
        best_route.append(next_state)
        state = next_state

    # print route in names
    route_names = []
    for i in best_route:
        route_names.append(statenames[i])
    print(best_route)
    print(route_names)
    print(q_vals)

    # run best route
    df1, initialerror = lr_get_mse(safecopy, "target")
    df2 = pd.DataFrame.copy(safecopy)
    finalerror = 0
    for i in range(0, len(best_route) - 1):
        df2, finalerror = statefuncs[best_route[i]](df2, target)
    print("MSE:", initialerror, "-->", finalerror)
    print("Diff:", initialerror - finalerror)
    return df2



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

with open("README.rst") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

requirements = [
    "numpy==1.21.4",
    "pandas==1.3.4",
    "scipy==1.7.3",
    "scikit-learn==1.0.1",
]

test_requirements = []

setup(
    author="William Su, Yash Modi, Niranjan Reddy Masapeta",
    author_email="modiyash3393@gmail.com",
    python_requires=">=3.6",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    description="A tool to automate preprocessing phase.",
    install_requires=requirements,
    license="MIT license",
    long_description=readme + "\n\n" + history,
    include_package_data=True,
    keywords="preprocmachine",
    name="preprocmachine",
    packages=find_packages(exclude=["tests"]),
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/yashm28sjsu/preprocmachine",
    version="0.0.2",
    zip_safe=False,
)
